/*
 * ==========================================================================================
 * check.c
 *
 * Date: 10 May 2004
 * Author: Jerome Delamarche
 * Version: 1.3
 *
 * ------------------------------------------------------------------------------------------
 * CHANGES:
 *
 * ------------------------------------------------------------------------------------------
 *
 * TODO: 
 *
 * ------------------------------------------------------------------------------------------
 *
 * Apply the rules !
 *
 * ==========================================================================================
 */
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include "common.h"
#include "rules.h"
#include "http_log.h"

static const char *rcsid = "$Id$";

static const char *g_IllegalParmEnvVar = "PARMGUARD_PARM_ILLEGAL";
static const char *g_UndefParmEnvVar   = "PARMGUARD_PARM_NOT_CHECKED";
static void (*perr)(const char *format,...);

/*
 * ------------------------------------------------------------------------------------------
 *
 * Common error generation:
 *
 * ------------------------------------------------------------------------------------------
 */
static int ConstraintError(request_rec *r, const char *uri, const char *parm, lsParmguardCtxt *ctxt)
{
	if (ctxt->illegal_parm_action_log) {
		(*perr)("Constraint mismatch for parameter '%s' of URI '%s'",parm,uri);
	}

	if (ctxt->illegal_parm_action_setenv) {
		/* add the environnemnt variable if not already done
 		 * (by a check with the server level conf file for example):
		 */
		char *vname = apr_pstrcat(r->pool,g_IllegalParmEnvVar,"_",parm,NULL);
		if (!apr_table_get(r->subprocess_env,vname)) {
			const char *value = apr_table_get(r->subprocess_env,g_IllegalParmEnvVar);
			apr_table_addn(r->subprocess_env,vname,"1");

			if (value) {
				char *newvalue = apr_pstrcat(r->pool,value,",",parm,NULL);
				apr_table_set(r->subprocess_env,g_IllegalParmEnvVar,newvalue);
			}
			else {
				apr_table_addn(r->subprocess_env,g_IllegalParmEnvVar,parm);
			}
		}
	}

	return ctxt->http_error_code;
}

static int UndefParmError(request_rec *r, const char *uri, const char *parm, lsParmguardCtxt *ctxt)
{
	if (ctxt->undef_parm_action_log) {
		(*perr)("No Constraint for parameter '%s' of URI '%s'",parm,uri);
	}

	if (ctxt->undef_parm_action_setenv) {
		const char *value = apr_table_get(r->subprocess_env,g_UndefParmEnvVar);
		char *vname = apr_pstrcat(r->pool,g_UndefParmEnvVar,"_",parm,NULL);
		apr_table_addn(r->subprocess_env,vname,"1");

		if (value) {
			char *newvalue = apr_pstrcat(r->pool,value,",",parm,NULL);
			apr_table_addn(r->subprocess_env,g_UndefParmEnvVar,newvalue);
		}
		else {
			apr_table_addn(r->subprocess_env,g_UndefParmEnvVar,parm);
		}
	}

	return ctxt->http_error_code;
}

/*
 * ==========================================================================================
 *
 * Apply the rules to the given URI:
 *
 * ==========================================================================================
 */
/*
 * ------------------------------------------------------------------------------------------
 *
 * Apply the string constraints:
 *
 * ------------------------------------------------------------------------------------------
 */
static int ApplyStringConstraints(request_rec *r, apr_uri_t *uri, guard_parm_t *parm, 
					guard_aval_t *attr, char *val, lsParmguardCtxt *ctxt)
{
	int invalid = 0;

	if (g_debuglevel)
		(*perr)("ApplyStringConstraints on parm '%s'",NAME(parm));

	while (attr) {
		switch (attr->m_attr->m_ct) {
		case CONSTRAINT_MAXLEN: invalid = (strlen(val) > attr->m_lvalue); break;
		case CONSTRAINT_MINLEN: invalid = (strlen(val) < attr->m_lvalue); break; 
		case CONSTRAINT_CHARCLASS: invalid = (regexec(attr->m_regexp,val,0,NULL,0)); break; 
		}

		if (invalid) {
			int ret = ConstraintError(r,uri->path,NAME(parm),ctxt);
			/* if (ctxt->illegal_parm_action_reject) */
				return ret;
		}

		attr = NEXT(attr);
	}

	return 0;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Apply the integer constraints:
 *
 * ------------------------------------------------------------------------------------------
 */
static int ApplyIntegerConstraints(request_rec *r, apr_uri_t *uri, guard_parm_t *parm, 
					guard_aval_t *attr, char *val, lsParmguardCtxt *ctxt)
{
	int invalid = 0;
	char *endptr;
	long value;

	if (g_debuglevel)
		(*perr)("ApplyIntegerConstraints on parm '%s'",NAME(parm));

	/* +++++ added by JD on the 27 Aug 2003:
	 * +++++ empty string ARE NOT considered as 0s !
	 */
	if (*val == '\0') {
		int ret = ConstraintError(r,uri->path,NAME(parm),ctxt);
		/* if (ctxt->illegal_parm_action_reject) */
			return ret;
	}

	while (attr) {
		if (g_debuglevel)
			(*perr)("Constraint '%d', val=%s, m_lvalue=%ld",attr->m_attr->m_ct,val,attr->m_lvalue);

		value = strtol(val,&endptr,10);

		/* +++++ modified by JD on the 27 Aug 2003:
		 * +++++ added errno == ERANGE condition
	         */
		/* if (value == LONG_MIN || value == LONG_MAX || *endptr) { */
		if (((value == LONG_MIN || value == LONG_MAX) && errno == ERANGE) || *endptr) {
			int ret = ConstraintError(r,uri->path,NAME(parm),ctxt);
			/* if (ctxt->illegal_parm_action_reject) */
				return ret;
		}

		switch (attr->m_attr->m_ct) {
		case CONSTRAINT_MINVAL: invalid = (value < attr->m_lvalue); break; 
		case CONSTRAINT_MAXVAL: invalid = (value > attr->m_lvalue); break; 
		}

		if (invalid) {
			int ret = ConstraintError(r,uri->path,NAME(parm),ctxt);
			/* if (ctxt->illegal_parm_action_reject) */
				return ret;
		}

		attr = NEXT(attr);
	}

	return 0;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Apply the decimal constraints:
 *
 * ------------------------------------------------------------------------------------------
 */
static int ApplyDecimalConstraints(request_rec *r, apr_uri_t *uri, guard_parm_t *parm, 
					guard_aval_t *attr, char *val, lsParmguardCtxt *ctxt)
{
	int invalid = 0;
	char *endptr;
	double value;

	if (g_debuglevel)
		(*perr)("ApplyDecimalConstraints on parm '%s'",NAME(parm));

	while (attr) {
		if (g_debuglevel)
			(*perr)("Constraint '%d', val=%s, m_fvalue=%lf",attr->m_attr->m_ct,val,attr->m_fvalue);

		value = strtod(val,&endptr); 
		if (errno || *endptr) { 
			int ret = ConstraintError(r,uri->path,NAME(parm),ctxt);
			/* if (ctxt->illegal_parm_action_reject) */
				return ret;
		}

		switch (attr->m_attr->m_ct) {
		case CONSTRAINT_MINVAL: invalid = (value < attr->m_fvalue); break; 
		case CONSTRAINT_MAXVAL: invalid = (value > attr->m_fvalue); break; 
		}

		if (invalid) {
			int ret = ConstraintError(r,uri->path,NAME(parm),ctxt);
			/* if (ctxt->illegal_parm_action_reject) */
				return ret;
		}

		attr = NEXT(attr);
	}

	return 0;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Apply the enum constraints:
 *
 * ------------------------------------------------------------------------------------------
 */
static int ApplyEnumConstraints(request_rec *r, apr_uri_t *uri, guard_parm_t *parm, 
					guard_aval_t *attr, char *val, lsParmguardCtxt *ctxt)
{
	int valid = 0;

	if (g_debuglevel)
		(*perr)("ApplyEnumConstraints on parm '%s'",NAME(parm));

	while (attr) {
		switch (attr->m_attr->m_ct) {
		case CONSTRAINT_OPTION: if (!strcmp(val,attr->m_svalue)) valid = 1; break; 
		}

		if (valid) break;

		attr = NEXT(attr);
	}

	if (!valid) {
		int ret = ConstraintError(r,uri->path,NAME(parm),ctxt);
		/* if (ctxt->illegal_parm_action_reject) */
			return ret;
	}

	return 0;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Apply the constraints to the specified paramater:
 *
 * ------------------------------------------------------------------------------------------
 */
static int ApplyConstraints(request_rec *r, apr_uri_t *uri, guard_parm_t *parm, char *val, lsParmguardCtxt *ctxt)
{
	guard_type_t *type = parm->m_type;
	guard_aval_t *attr = parm->m_value;
	int status = 0;

	if (g_debuglevel)
		(*perr)("ApplyConstraints on parm '%s'",NAME(parm));

	/* base type or user defined one ? */
	if (type->m_userdefined) {
		attr = type->m_value;
		type = type->m_basetype;
	}

	/* Manage with the predefined types: */
	switch (type->m_bt) {
	case CONSTRAINT_STRING : status = ApplyStringConstraints(r,uri,parm,attr,val,ctxt); break;
	case CONSTRAINT_INTEGER: status = ApplyIntegerConstraints(r,uri,parm,attr,val,ctxt); break;
	case CONSTRAINT_ENUM   : status = ApplyEnumConstraints(r,uri,parm,attr,val,ctxt); break;
	case CONSTRAINT_DECIMAL: status = ApplyDecimalConstraints(r,uri,parm,attr,val,ctxt); break;
	}

	return status;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Apply a rule to the request:
 *
 * ------------------------------------------------------------------------------------------
 */
static int ApplyRule(request_rec *r, guard_rule_t *rule, apr_uri_t *uri, lsParmguardCtxt *ctxt)
{
	guard_parm_t 	*parm;
	char *q = uri->query;
	char *end,*equal;
	int status;
	int reterror = 0;
	char *copy = NULL, *pcopy = NULL;

	assert(q);

	/* unescape the %xx characters: */
	if (ap_unescape_url(q))
		return 0;	/* no error generated */

	if (g_debuglevel) (*perr)("query=%s",q); 

	/* if "clear" or "filter" action specified, work on a copy: */
	if (ctxt->illegal_parm_action_clear || ctxt->illegal_parm_action_filter ||
	    ctxt->undef_parm_action_clear) {
		pcopy = copy = apr_palloc(r->pool,strlen(q)+1);
	}

	/* Split the query to get the pairs of parms/values */
	/* +++++ fixed on the 11 Aug 2003 */
	/* +++++ v1.0: while (*q) { */
	while (q && *q) {
		end = strchr(q,'&');
		if (end) *end = '\0';

		equal = strchr(q,'=');
		if (!equal) 
			goto nextparm;

		*equal = '\0';	/* q points to the parm and equal+1 points to the value */
		if (copy) { strcpy(pcopy,q); pcopy += strlen(q); *pcopy++ = '='; }

		/* Note: empty parms are handled like unknown parms */

		/* Is there a description for this parm ? */
		if (parm = SearchParm(rule,q)) {
			if (status = ApplyConstraints(r,uri,parm,equal+1,ctxt)) {
				/* clear the parameter (do not copy it !) if required: */
				if (copy && !ctxt->illegal_parm_action_clear) {
					strcpy(pcopy,equal+1); 
					pcopy += strlen(equal+1);
				}

				if (ctxt->illegal_parm_action_reject) {
					if (!ctxt->scan_all_parm)
						return status;
					reterror = status;
				}
			}
			else if (copy) {
				strcpy(pcopy,equal+1); 
				pcopy += strlen(equal+1);
			}
		}
		else {
			/* handling of unknown parameters: */
			if (copy && !ctxt->undef_parm_action_clear) {
				strcpy(pcopy,equal+1); 
				pcopy += strlen(equal+1);
			}

			if (ctxt->undef_parm_action_reject) {
				UndefParmError(r,r->uri,q,ctxt);
				if (!ctxt->scan_all_parm)
					return ctxt->http_error_code;
				reterror = status;
			}
		}

		*equal = '=';	/* restore the initial query */

	nextparm:
		if (copy && !equal) { strcpy(pcopy,q); pcopy += strlen(q); }
		/* next parm: */
		q = (end)?end+1:NULL;

		/* restore the initial query !!! */
		if (end) {
			if (copy) *pcopy++ = '&';
			*end = '&';
		}
	}

	if (copy) {
		*pcopy = '\0';
		uri->query = copy;
	}
	if (g_debuglevel) (*perr)("copy=%s",copy); 

	return reterror;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Main function. Called for each request.
 *
 * ------------------------------------------------------------------------------------------
 */
int ApplyRules(void (*errfunc)(const char *format,...), guard_ctxt_t *cfg, request_rec *r)
{
	apr_uri_t *uri = &r->parsed_uri;
	int one_matched = 0;
	lsParmguardCtxt *ctxt; 
	guard_rule_t *rule;

	assert(errfunc);
	perr = errfunc;

	/* no rules ? */
	if (!cfg) return 0;

	if (g_debuglevel)
		(*perr)("Apply Rules of file '%s' [stored at 0x%x]",cfg->conffile,cfg);

	/* Ignore requests without args: */
	if (!uri->query || !*(uri->query))
		return 0;

	/* Loop on the rules: */
	rule = GetRuleList(cfg);
	ctxt = &cfg->global_ctxt;

	while (rule) {
		/* URL match ? */
		if (!regexec(rule->m_regexp,uri->path,0,NULL,0)) {
			int status;
			one_matched = 1;

			if (g_debuglevel)
				(*perr)("Rule match with'%s'",rule->m_path);

			if (status = ApplyRule(r,rule,uri,ctxt))
				return status;
		}

		/* many rules may match the URI */
		rule = NEXT(rule);
	}

	/* Check if URL has been 'parmguarded': */
	if (!one_matched) {
		if (ctxt->undef_url_action_log) {
			(*perr)("Warning: Request '%s' is not parmguarded",uri->path);
		}

		if (ctxt->undef_url_action_setenv) {
			apr_table_addn(r->subprocess_env,"PARMGUARD_URL_NOT_CHECKED","1");
		}

		if (ctxt->undef_url_action_reject) {
			if (ctxt->undef_url_action_log) {
				(*perr)("Request '%s' rejected because it is not 'parmguarded'",uri->path);
			}

			return ctxt->http_error_code;
		}
	}

	return 0;
}

/*
 * $Log:$
 */

/* EOF */
