/*
 * ==========================================================================================
 * conf.c
 *
 * Date: 10 May 2004 
 * Author: Jerome Delamarche
 * Version: 1.3 
 *
 * ------------------------------------------------------------------------------------------
 * CHANGES:
 *
 * ------------------------------------------------------------------------------------------
 *
 * TODO:
 *
 * ------------------------------------------------------------------------------------------
 *
 * Manages with the XML configuration files
 *
 * ==========================================================================================
 */
#include <assert.h>

#include "common.h"
#include "conf.h"
#include "http_log.h"

static const char *rcsid = "$Id$";
static void (*perr)(const char *format,...);

/* Module global variables */
#define MAX_CONF_FILE	100

/* These variables are initialized at startup, no mutex needed */
static conf_file_t 	g_ConfFile[MAX_CONF_FILE];
static int 		g_ConfNu = 0;
static int		g_ConfMax = MAX_CONF_FILE; 

/*
 * ------------------------------------------------------------------------------------------
 *
 * Initializes the XML library
 *
 * ------------------------------------------------------------------------------------------
 */
void InitConf(void (*errfunc)(const char *format,...))
{
	if (!*errfunc) exit(99);
	perr = errfunc;

	xmlInitParser();
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Frees the XML library
 *
 * ------------------------------------------------------------------------------------------
 */
void EndConf(void)
{
	xmlCleanupParser();
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Check if the given file has already been parsed:
 *
 * ------------------------------------------------------------------------------------------
 */
conf_file_t *CheckFileConf(const char *filename)
{
	int i;
	conf_file_t *pf;

	for ( i = 0, pf = g_ConfFile ; i < g_ConfNu ; i++,pf++ ) {
		if (!strcmp(pf->filename,filename)) {
			return &g_ConfFile[i];
		}
	}

	return NULL;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Analyzes a configuration file
 *
 * ------------------------------------------------------------------------------------------
 */
xmlDocPtr ReadConfFile(const char *fname)
{
	xmlDocPtr	p_xmlDoc;

	assert(perr);
	if (!(p_xmlDoc = xmlParseFile(fname))) {
		(*perr)("Error while reading configuration file: %s",fname);
		return NULL;
	}

	return p_xmlDoc;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Store a file description:
 *
 * ------------------------------------------------------------------------------------------
 */
conf_file_t *StoreFileConf(const char *fname, guard_ctxt_t *ctxt)
{
	struct stat 	buf;
	int		filenu;

	/* ASSERT: The filename must not be already in the array: */

	if (g_ConfNu == g_ConfMax) {
		(*perr)("Configuration file table full: could not store file '%s'",fname);
		return NULL;
	}

	/* get the last modification time of the conf file: */
	if (!stat(fname,&buf)) {
		filenu = g_ConfNu;
		g_ConfFile[g_ConfNu].last_mtime = buf.st_mtime;
		g_ConfFile[g_ConfNu].filename   = fname;
		UpdateFileConf(&g_ConfFile[g_ConfNu],ctxt);
		g_ConfNu++;
		/* (*perr)("Last time =%d",buf.st_mtime); */
	}
	else {
		(*perr)("stat failed for file '%s': Could not get the time of last modification",fname);
		return NULL;
	}

	return &g_ConfFile[filenu];
}

void UpdateFileConf(conf_file_t *conffile, guard_ctxt_t *ctxt)
{
	conffile->rules = ctxt;
}

guard_ctxt_t *GetFileConf(conf_file_t *conffile)
{
	return conffile->rules;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Check the last modification time of the conf files:
 *
 * ------------------------------------------------------------------------------------------
 */
void CheckLastTimeConf(void (*updateConf)(conf_file_t *conffile))
{
	int i;
	struct stat buf;
	conf_file_t *pf;

	assert(perr);

	for ( i = 0, pf = g_ConfFile ; i < g_ConfNu ; i++,pf++ ) {
		if (!stat(pf->filename,&buf)) {
			if (buf.st_mtime > pf->last_mtime) {
				(*updateConf)(pf);
				pf->last_mtime = buf.st_mtime;
			}
		}
		else {
			(*perr)("stat failed for file %s [%s]",pf->filename,strerror(errno));
		}
	}
}

/*
 * $Log:$
 */

/* EOF */
