/*
 * ==========================================================================================
 * mod_parmguard.c
 *
 * Date: 10 May 2004 
 * Author: Jerome Delamarche
 * Version: 1.3 
 *
 * ------------------------------------------------------------------------------------------
 * CHANGES:
 *
 * ------------------------------------------------------------------------------------------
 *
 * TODO: add a lock when the conf changes !
 *
 * ------------------------------------------------------------------------------------------
 *
 * Example usage:
 *
 * LoadModule parmguard_module /apache/dir/mod_parmguard.so
 * AddModule mod_parmguard.c
 *
 * ==========================================================================================
 */
#include "common.h"

#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#if WITH_APACHE2
#include "apr_thread_rwlock.h"
#endif
#include "scoreboard.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/un.h>

#include "conf.h"
#include "rules.h"

static const char *rcsid = "$Id$";

/* 
 * private configuration: 
 */
typedef struct {
	int				isOn;		/* directive ParmguardEngine on|off */
	int				traceDebug;	/* debug level */
	conf_file_t			*conffile;
} guard_config_t;

/* Module private and VERY GLOBAL variables */
static int g_PeriodicReload = 0;	/* time (in seconds) between re-reading the conf file (0 means never) */
static int g_TraceDebug = 0;
#if WITH_APACHE2
static apr_thread_rwlock_t *g_rwlock;
#endif

/* 
 * Forward declaration... 
 */
module parmguard_module;

/*
 * --------------
 * Error Callback: 
 * --------------
 */
static void ParmguardError(const char *format, ...)
{
	char buffer[512];
	va_list ap;
	va_start(ap,format);
	vsnprintf(buffer,sizeof(buffer)-1,format,ap);

	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)NULL,buffer); 
	va_end(ap);
}

/*
 * ------------------------------
 * Standard Apache function calls:
 * ------------------------------
 */
void *APR_PALLOC(void *p, unsigned int t) { return apr_palloc(p,t); }
void *APR_PSTRDUP(void *p, const char *s) { return apr_pstrdup(p,s); }
ap_regex_t *AP_PREGCOMP(void *p, const char *pat, int fl) { return ap_pregcomp(p,pat,fl); }
apr_status_t APR_POOL_CREATE(apr_pool_t **p) { return (apr_status_t)apr_pool_create(p,NULL); }
void APR_POOL_DESTROY(apr_pool_t *p) { apr_pool_destroy(p); }

#if WITH_APACHE2
/* 
 * ---------------------------------------------------------------
 * Handling of the timer used to reload the conf files dynamically:
 * ---------------------------------------------------------------
 */
/* This is a callback invoked by CheckLastTimeConf() for each modified conf file: */
static void UpdateConf(conf_file_t *conffile)
{
	xmlDocPtr	pDoc;
	guard_ctxt_t	*ctxt;
	const char	*fname = conffile->filename;

	if (g_TraceDebug)
		ParmguardError("UpdateConf: %s",fname);

	/* 1-Reload the new file and check for error: */
	InitConf(ParmguardError);
	if (!(pDoc = ReadConfFile(fname))) {
		EndConf();
		ParmguardError("File '%s' changed but its syntax is invalid",fname);
		return;
	}

	/* Check the conf XML file semantics: */
	ctxt = InitRules(g_TraceDebug,fname,ParmguardError);
	if (!ctxt || BuildRules(ctxt,pDoc)) {
		ParmguardError("File '%s' changed but its content is invalid",fname);
		EndRules(ctxt);
		EndConf();
		return;
	}

	EndConf();	/* necessary to free the memory allocated by libxml */
	/* if (g_TraceDebug) DumpRules(ctxt); */

	/* 2-set a lock to be able to switch the rules: */
	/* ParmguardError("before rwlock() by %d",getpid()); */
	if (apr_thread_rwlock_wrlock(g_rwlock))
		ParmguardError("Error while trying to lock the rwlock");

	/* 3-change the rules and free the old rules: */
	if (g_TraceDebug)
		ParmguardError("new ctxt=0x%x for conffile=0x%x",ctxt,conffile);

	EndRules(GetFileConf(conffile));
	UpdateFileConf(conffile,ctxt);

	/* 4-unlock the mutex: */
	if (apr_thread_rwlock_unlock(g_rwlock))
		ParmguardError("Error while trying to unlock the rwlock");
}

static void TimerHandler(int nouse) 
{
	/*
	if (g_TraceDebug)
		ParmguardError("TimerHandler");
	*/

	CheckLastTimeConf(UpdateConf);

	alarm(g_PeriodicReload);
}

static void TimerStart()
{
	/* we don't care if an alarm was already set */
	alarm(g_PeriodicReload);
	signal(SIGALRM,TimerHandler);
}

static void TimerStop()
{
	signal(SIGALRM,SIG_IGN);
}
#endif

/* 
 * ------------------
 * Directive handlers: 
 * ------------------
 */

#if WITH_APACHE2
static apr_status_t parmguard_cleanup(void *v)
#else
static void parmguard_cleanup(void *v)
#endif
{
#if WITH_APACHE2
	/* ParmguardError("lock destroyed by %d",getpid()); */
	if (apr_thread_rwlock_destroy(g_rwlock))
		ParmguardError("Unable to destroy the rwlock");

	return APR_SUCCESS;
#endif
}

#if WITH_APACHE2
static int parmguard_init(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
#endif
#if WITH_APACHE1
static void parmguard_init(server_rec *s, apr_pool_t *p)
#endif
{
	ParmguardError("parmguard_init");

#if WITH_APACHE2
	if (apr_thread_rwlock_create(&g_rwlock,p))
		ParmguardError("Could not create the thread mutex");
	/* apr_register_cleanup(p,NULL,parmguard_cleanup,apr_pool_cleanup_null); */

	return APR_SUCCESS;
#endif
}

#if WITH_APACHE2
/*
 * Hook for child init: 
 */
static void parmguard_child_init(apr_pool_t *p, server_rec *s)
{
	/* ParmguardError("lock created by %d",getpid()); */
	if (apr_thread_rwlock_create(&g_rwlock,p))
		ParmguardError("Could not create the thread mutex");

	/* initialize the alarm: */
	TimerStart();
}
#endif

/*
 * ----------------------------
 * Directives at Location Level:
 * ----------------------------
 */
static const char *parmguard_engine_cmd(cmd_parms *parms, void *config, int bool)
{
	guard_config_t *cfg = (guard_config_t *)config;
	cfg->isOn = bool;

	return NULL;
}

/*
 * --------------------------
 * Directives at Server Level:
 * --------------------------
 */
static const char *parmguard_trace_cmd(cmd_parms *parms, void *config, char *cmd)
{
	guard_config_t *cfg = (guard_config_t *)ap_get_module_config(parms->server->module_config,&parmguard_module);

	if (strstr(cmd,"debug")) {
		g_TraceDebug = cfg->traceDebug = 1;
	}
	else {
		char *msg = "Invalid value for ParmguardTrace directive";
		ParmguardError("%s: %s",msg,cmd); 
		return msg;
	}

	return NULL;
}

static const char *parmguard_periodic_reload(cmd_parms *parms, void *config, char *cmd)
{
	guard_config_t *cfg = (guard_config_t *)ap_get_module_config(parms->server->module_config,&parmguard_module);
	int val = atoi(cmd);

#if WITH_APACHE1
	char *notsup = "Directive supported with Apache 2.x only";
	ParmguardError(notsup);
	return notsup;
#endif

	if (val >= 0) {
		g_PeriodicReload = val;
		TimerStart();
	}
	else {
		char *msg = "Invalid value for ParmguardPeriodicReload directive";
		ParmguardError("%s: %s",msg,cmd); 
		return msg;
	}

	return NULL;
}

/*
 * -------------------------------------
 * Directives at Server & Location Level:
 * -------------------------------------
 */
static const char *parmguard_conffile_cmd(cmd_parms *parms, void *config, char *cmd)
{
	guard_config_t 	*cfg;
	guard_config_t 	*srvcfg;
	xmlDocPtr	pDoc;
	char 		*s;
	guard_ctxt_t	*ctxt;

	srvcfg = (guard_config_t *)ap_get_module_config(parms->server->module_config,&parmguard_module);

	/* called for at Location or a Server Context ? */
	if (parms->path) {
		cfg = (guard_config_t *)config;

		if (srvcfg->traceDebug)
			ParmguardError("conffile_cmd for path: %s",parms->path);
	}
	else {
		cfg = srvcfg;

		if (srvcfg->traceDebug)
			ParmguardError("conffile_cmd for server");
	}

	/* Check if the file has already been parsed: */
	if (cfg->conffile = CheckFileConf(cmd)) {
		if (srvcfg->traceDebug)
			ParmguardError("File %s already loaded",cmd);
		return NULL;	
	}

	/* First time parse: */
	InitConf(ParmguardError);
	if (!(pDoc = ReadConfFile(cmd))) {
		EndConf();
		return "Configuration Error";
	}

	/* Check the conf XML file semantics: */
	ctxt = InitRules(srvcfg->traceDebug,cmd,ParmguardError);
	if (!ctxt || BuildRules(ctxt,pDoc)) {
		EndRules(ctxt);
		EndConf();
		return "Incorrect Rules";
	}

	cfg->conffile = StoreFileConf(cmd,ctxt);
	if (g_TraceDebug)
		ParmguardError("new ctxt=0x%x for conffile=0x%x",GetFileConf(cfg->conffile),cfg->conffile);
		
	EndConf();	/* necessary to free the memory allocated by libxml */

	/* if (srvcfg->traceDebug) DumpRules(ctxt); */

	return NULL;
}

/* 
 * ------------------------
 * Module Apache directives:
 * ------------------------
 */
command_rec parmguard_cmds[] = {
    AP_INIT_FLAG("ParmguardEngine", parmguard_engine_cmd, NULL, 
     ACCESS_CONF, "Start/Stop the engine"),
    AP_INIT_TAKE1("ParmguardTrace", parmguard_trace_cmd, NULL, 
     RSRC_CONF, "Debug listed components (debug)"),
    AP_INIT_TAKE1("ParmguardConfFile", parmguard_conffile_cmd, NULL, 
     RSRC_CONF|ACCESS_CONF, "Configuration file"),
    AP_INIT_TAKE1("ParmguardPeriodicReload", parmguard_periodic_reload, NULL, 
     RSRC_CONF, "Time (in seconds) before re-reading the configuration files"),
    {NULL}
};

/*
 * --------------------------------
 * Directory Configuration Callback:
 * --------------------------------
 */
static void *create_dir_config(apr_pool_t *p, char *path)
{
	guard_config_t *cfg = (guard_config_t *)apr_pcalloc(p,sizeof(guard_config_t));

	memset(cfg,0,sizeof(guard_config_t));

	return (void *)cfg;
}

/*
 * --------------------------------
 * Directory Configuration Callback:
 * --------------------------------
 */
static void *create_server_config(apr_pool_t *p, server_rec *r)
{
	guard_config_t *cfg = (guard_config_t *)apr_pcalloc(p,sizeof(guard_config_t));

	memset(cfg,0,sizeof(guard_config_t));

	return (void *)cfg;
}

/*
 * ------------------------------
 * Filtering function:
 * ------------------------------
 */
static int parmguard_start(request_rec *r)
{
	struct timeval 	stamp;
	char 		buff[64];
	int 		idx;
	guard_config_t 	*scf = (guard_config_t *)ap_get_module_config(r->server->module_config,&parmguard_module);
	guard_config_t 	*cfg = (guard_config_t *)ap_get_module_config(r->per_dir_config,&parmguard_module);
	apr_uri_t    	*uri = &r->parsed_uri;
	int 		status;

	if (scf->traceDebug) 
		ParmguardError("--> parmguard_start on URI: %s?%s, method=%d, scf=0x%x, cfg=0x%x conffile1=%s, conffile2=%s",
			uri->path,uri->query,r->method_number,scf,cfg,
			(scf && scf->conffile)?scf->conffile->filename:"",
			(cfg && cfg->conffile)?cfg->conffile->filename:"");

	/* Configured at Location or Server level ? */
	if (!cfg) cfg = scf;

	/* parmguarding is on ? */
	if (!cfg->isOn) {
		if (scf->traceDebug)
			ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,ALE2(APR_SUCCESS)r->server,"<-- engine off");
		return DECLINED;
	}

	/* the following code is taken and slighlty adapted from mod_cgi.c: */
	if (r->method_number == M_POST) {
#if WITH_APACHE2
		apr_bucket_brigade *bb;
		int seen_eos = 0;
		apr_status_t rv;
		uri->query = "";

		bb = apr_brigade_create(r->pool,r->connection->bucket_alloc);
		do {
			apr_bucket *bucket;
			rv = ap_get_brigade(r->input_filters,bb,AP_MODE_READBYTES,APR_BLOCK_READ,HUGE_STRING_LEN);
			if (rv != APR_SUCCESS)
				return rv;

		        for (bucket = APR_BRIGADE_FIRST(bb);
		             bucket != APR_BRIGADE_SENTINEL(bb);
		             bucket = APR_BUCKET_NEXT(bucket)) {
				const char *data;
				apr_size_t len;

				if (APR_BUCKET_IS_EOS(bucket)) {
					seen_eos = 1;
					break;
				}

				if (APR_BUCKET_IS_FLUSH(bucket)) {
					continue;
				}

				apr_bucket_read(bucket,&data,&len,APR_BLOCK_READ);

				uri->query = apr_pstrcat(r->pool,uri->query,data,NULL);
			}
			apr_brigade_cleanup(bb);
		} while (!seen_eos);
#endif

#if WITH_APACHE1
		uri->query = "";

		if (status = ap_setup_client_block(r,REQUEST_CHUNKED_ERROR))
			return status;

		if (ap_should_client_block(r)) {
			int len_read;
			char argsbuffer[HUGE_STRING_LEN];

			ap_hard_timeout("copy POST args",r);

			while ((len_read = ap_get_client_block(r,argsbuffer,HUGE_STRING_LEN)) > 0) {
				uri->query = apr_pstrcat(r->pool,uri->query,argsbuffer,NULL);
				ap_reset_timeout(r);
			}

			ap_kill_timeout(r);
		}
#endif

		/* Simulate a GET: */
		r->method_number = M_GET;
		/* r->unparsed_uri  = apr_pstrcat(r->pool,r->uri,"?",r->parsed_uri.query,NULL); */
		r->args = r->parsed_uri.query;

		if (scf->traceDebug)
			ParmguardError("data=%s",r->parsed_uri.query);
	}

	/* check the local rules first, then the global ones (server level): */
	if (scf->traceDebug)
		ParmguardError("cfg->conffile=0x%x rules=0x%x), scf->conffile=0x%x",cfg->conffile,cfg->conffile->rules,scf->conffile);

	status = 0;

#if WITH_APACHE2
	if (apr_thread_rwlock_rdlock(g_rwlock))
		ParmguardError("Error while trying to lock the rwlock in read mode");
#endif

	if (cfg->conffile) status = ApplyRules(ParmguardError,GetFileConf(cfg->conffile),r);
	if (!status && scf->conffile) status = ApplyRules(ParmguardError,GetFileConf(scf->conffile),r);

#if WITH_APACHE2
	if (apr_thread_rwlock_unlock(g_rwlock))
		ParmguardError("Error while trying to unlock the rwlock");
#endif
	
	if (status) {
		/* apr_table_addn(r->subprocess_env,"BAD_PARM","1"); */
		if (scf->traceDebug)
			ParmguardError("<-- status=%d",status);

		return status;
	}

	if (scf->traceDebug)
		ParmguardError("<-- OK");

    	return DECLINED;
}

#if WITH_APACHE2
/*
 * ------------------
 * Hooks Registration:
 * ------------------
 */
static void register_hooks(apr_pool_t *pool)
{
	/*
	ParmguardError("register mod_parmguard hooks");
	*/
	ap_hook_post_config(parmguard_init,NULL,NULL,APR_HOOK_LAST);
	/*
	 * Modified by JD on the 12 Feb 2003:
	 * APR_HOOK_FIRST instead of LAST ! (fix a bug in Apache 2.0.44+)
	 */
	ap_hook_fixups(parmguard_start,NULL,NULL,APR_HOOK_REALLY_FIRST);
	ap_hook_child_init(parmguard_child_init,NULL,NULL,APR_HOOK_MIDDLE);
}

/*
 * -----------------
 * Module Definition:
 * -----------------
 */
module AP_MODULE_DECLARE_DATA parmguard_module = {
    STANDARD20_MODULE_STUFF,
    create_dir_config,			/* dir config creater */
    NULL,				/* dir merger default is to override */
    create_server_config,		/* server config */
    NULL,				/* merge server config */
    parmguard_cmds,			/* command table */
    register_hooks
};
#endif

#if WITH_APACHE1
module MODULE_VAR_EXPORT parmguard_module = {
    STANDARD_MODULE_STUFF,
    parmguard_init,			/* initializer */
    create_dir_config,			/* dir config creater */
    NULL,				/* dir merger default is to override */
    create_server_config,		/* server config */
    NULL,				/* merge server config */
    parmguard_cmds,			/* command table */
    NULL,				/* handlers */
    parmguard_start,			/* filename translation */
    NULL,				/* check_user_id */
    NULL,				/* check auth */
    NULL,				/* check access */
    NULL,				/* type_checker */
    NULL,				/* fixups */
    NULL,				/* logger */
    NULL,				/* header parser */
    NULL,				/* child_init */
    NULL,				/* child_exit */
    NULL				/* post read-request */
};
#endif

/*
 * $Log:$
 */

/* EOF */
