/*
 * ==========================================================================================
 * rules.c
 *
 * Date: 10 May 2004 
 * Author: Jerome Delamarche
 * Version: 1.3 
 *
 * ------------------------------------------------------------------------------------------
 * CHANGES:
 *
 * ------------------------------------------------------------------------------------------
 *
 * TODO: define predefined types only once
 *
 * ------------------------------------------------------------------------------------------
 *
 * Manages with the constraint rules
 *
 * ==========================================================================================
 */
#include <assert.h>
#include <stdlib.h>
#include "common.h"
#include "rules.h"
#include "http_log.h"

static const char *rcsid = "$Id$";

/*
 * Global variables:
 */
/*
static guard_rule_t	*gp_guard_rule = NULL;
static guard_type_t	*gp_guard_type = NULL;
*/

int	g_debuglevel	  = 0;
static void (*perr)(const char *format,...);

/*
 * ------------------------------------------------------------------------------------------
 *
 * Initializes the rules with the predefined values:
 *
 * ------------------------------------------------------------------------------------------
 */
guard_ctxt_t *InitRules(int debuglevel, const char *filename, 
			void (*errfunc)(const char *format,...))
{
	apr_pool_t	*pool = NULL;
	guard_type_t	*gtype;
	guard_ctxt_t	*ctxt; 

	apr_status_t 	status = APR_POOL_CREATE(&pool);

	if (!*errfunc) exit(99);
	perr = errfunc;

	g_debuglevel = debuglevel;

	ctxt = APR_PALLOC(pool,sizeof(guard_ctxt_t));
	memset(ctxt,0,sizeof(guard_ctxt_t));

	ctxt->pool					= pool;
	ctxt->conffile					= filename;
	ctxt->global_ctxt.http_error_code 		= HTTP_NOT_ACCEPTABLE;
	ctxt->global_ctxt.undef_url_action_reject 	= 1;
	ctxt->global_ctxt.undef_url_action_log 		= 1;
	ctxt->global_ctxt.undef_url_action_setenv 	= 1;
	ctxt->global_ctxt.illegal_parm_action_reject 	= 1;
	ctxt->global_ctxt.illegal_parm_action_clear 	= 0;
	ctxt->global_ctxt.illegal_parm_action_filter 	= 0;
	ctxt->global_ctxt.illegal_parm_action_log 	= 1;
	ctxt->global_ctxt.illegal_parm_action_setenv 	= 1;
	ctxt->global_ctxt.undef_parm_action_reject 	= 1;
	ctxt->global_ctxt.undef_parm_action_clear 	= 0;
	ctxt->global_ctxt.undef_parm_action_log 	= 1;
	ctxt->global_ctxt.undef_parm_action_setenv 	= 1;
	ctxt->global_ctxt.scan_all_parm 		= 0;

	/* TODO: The types are redefined for each Conf file. This could be enhanced... */

	/* Create the default type with their attributes: */
	/* STRING type: */
	gtype = AddBaseType(ctxt,"string",CONSTRAINT_STRING);
	AddTypeAttr(pool,gtype,"minlen",CONSTRAINT_MINLEN);
	AddTypeAttr(pool,gtype,"maxlen",CONSTRAINT_MAXLEN);
	AddTypeAttr(pool,gtype,"charclass",CONSTRAINT_CHARCLASS);

	/* INTEGER type: */
	gtype = AddBaseType(ctxt,"integer",CONSTRAINT_INTEGER);
	AddTypeAttr(pool,gtype,"minval",CONSTRAINT_MINVAL);
	AddTypeAttr(pool,gtype,"maxval",CONSTRAINT_MAXVAL);

	/* DECIMAL type: */
	gtype = AddBaseType(ctxt,"decimal",CONSTRAINT_DECIMAL);
	AddTypeAttr(pool,gtype,"minval",CONSTRAINT_MINVAL);
	AddTypeAttr(pool,gtype,"maxval",CONSTRAINT_MAXVAL);

	/* ENUM type: */
	gtype = AddBaseType(ctxt,"enum",CONSTRAINT_ENUM);
	AddTypeAttr(pool,gtype,"multiple",CONSTRAINT_MULTIPLE);
	AddTypeAttr(pool,gtype,"option",CONSTRAINT_OPTION);

	xmlDoValidityCheckingDefaultValue = 1;

	return ctxt;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Frees the rules structures
 *
 * ------------------------------------------------------------------------------------------
 */
void EndRules(guard_ctxt_t *ctxt)
{
	APR_POOL_DESTROY(ctxt->pool);
}

/*
 * ==========================================================================================
 *
 * Build the rules from the given XML document:
 * (libxml should have validated the document using the specified DTD)
 *
 * ==========================================================================================
 */
/*
 * ------------------------------------------------------------------------------------------
 *
 * Manage with the constraints:
 *
 * ------------------------------------------------------------------------------------------
 */
static int BuildConstraintRule(guard_ctxt_t *ctxt, xmlNodePtr consNode, 
				guard_type_t **m_type, guard_aval_t **m_value)
{
	xmlNodePtr		child;
	guard_type_attr_t	*attr;
	xmlChar			*aname,*avalue;

	/* Loop on the children: */
	child = consNode->xmlChildrenNode;
	while (child) {
		if (child->type != XML_ELEMENT_NODE) {
			child = child->next;
			continue;
		}

		if (!xmlStrcmp(child->name,(const xmlChar *)"type")) {
			xmlChar *name = xmlGetProp(child,"name");
			guard_type_t *type;
			assert(!(*m_type));	/* only on type - DTD assumed */
			assert(name);		/* checked by DTD */
			if (!(type = SearchType(ctxt->p_guard_type,name))) {
				(*perr)("Type '%s' unknown",name);
				return 1;
			}
			*m_type = type;

			child = child->next;
			continue;
		}

		assert(*m_type);	/* type must have been set - DTD assumed */

		/* for yet, only <attr> node are allowed: */
		assert(!xmlStrcmp(child->name,(const xmlChar *)"attr"));

		aname = xmlGetProp(child,"name");
		avalue = xmlGetProp(child,"value");
		assert(aname && avalue);	/* DTD assumed */

		/* Check the relationship between type and attribute:
		 * if the type is a user type, the function performs a checking with the base type
		 */
		attr = SearchAttr(*m_type,(const char *)aname);
		if (!attr) {
			(*perr)("Constraint '%s' is not associated with type '%s'",aname,NAME(*m_type));
			return 1;
		}

		if (!xmlStrcmp(aname,(const xmlChar *)"maxlen") ||
		    !xmlStrcmp(aname,(const xmlChar *)"minlen") ||
		    !xmlStrcmp(aname,(const xmlChar *)"multiple") ||
		    !xmlStrcmp(aname,(const xmlChar *)"option") ||
		    !xmlStrcmp(aname,(const xmlChar *)"minval") ||
		    !xmlStrcmp(aname,(const xmlChar *)"maxval")) {
			AddValue(ctxt->pool,m_value,attr,avalue);
		}
		else if (!xmlStrcmp(aname,(const xmlChar *)"charclass")) {
			/* compile the regexp: */
			guard_aval_t *aval = AddValue(ctxt->pool,m_value,attr,avalue);
			ap_regex_t *reg = NULL;
			if (!(reg = AP_PREGCOMP(ctxt->pool,avalue,AP_REG_EXTENDED|AP_REG_NOSUB))) {
				(*perr)("Bad RE for 'charclass' constraint: %s",avalue);
				return 1;
			}
			aval->m_regexp = reg;
		}
		else {
			(*perr)("Constraint '%s' ignored",aname);
		}

		child = child->next;
	}

	return 0;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Manage with the <parm> node:
 *
 * ------------------------------------------------------------------------------------------
 */
static int BuildParmRule(guard_ctxt_t *ctxt, xmlNodePtr parmNode, guard_rule_t *rule)
{
	guard_parm_t	*parm = NULL;
	xmlNodePtr	child;
	xmlChar *name = xmlGetProp(parmNode,"name");
	int		status;

	assert(name);
	assert(rule);	/* <parm> must follow <match> */
	parm = AddParm(ctxt->pool,rule,(const char *)name);

	status = BuildConstraintRule(ctxt,parmNode,&parm->m_type,&parm->m_value);
	if (status) return status;

	return 0;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Manage with the <usertype> node:
 *
 * ------------------------------------------------------------------------------------------
 */
static int BuildUserType(guard_ctxt_t *ctxt, xmlDocPtr pDoc, xmlNodePtr typeNode)
{
	guard_type_t *type = NULL;
	xmlChar *name = xmlGetProp(typeNode,"name");
	xmlNodePtr	child;
	int		status;
	assert(name);

	/* Type must be unique: */
	if (SearchType(ctxt->p_guard_type,name)) {
		(*perr)("Type '%s' already defined",name);
		return 1;
	}

	/* Create the new type: */
	type = AddUserType(ctxt,name);

	status = BuildConstraintRule(ctxt,typeNode,&type->m_basetype,&type->m_value);
	if (status) return status;

	return 0;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Manage with the <url> node:
 *
 * ------------------------------------------------------------------------------------------
 */
static int BuildURLRule(guard_ctxt_t *ctxt, xmlDocPtr pDoc, xmlNodePtr cur)
{
	guard_rule_t *rule = NULL;
	xmlNodePtr child = cur->xmlChildrenNode;
	const char *match = NULL;

	while (child) {
		if (child->type != XML_ELEMENT_NODE) {
			child = child->next;
			continue;
		}

		if (!xmlStrcmp(child->name,(const xmlChar *)"match")) {
			match = (const char *)xmlNodeListGetString(pDoc,child->xmlChildrenNode,1);
			assert(!rule);	/* only one <match> */

			/* IMPROVE: server and port are not handled */
			rule = AddRule(ctxt,(const char *)NULL,(int)NULL,match);
			child = child->next;
			continue;
		}

		assert(rule);

		if (!xmlStrcmp(child->name,(const xmlChar *)"parm")) {
			int status = BuildParmRule(ctxt,child,rule);
			if (status) return status;
		}
		else {
			(*perr)("Unknown tag: %s",child->name);
			return 1;
		}

		child = child->next;
	}

	assert(match);

	return 0;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Manage with the <global> node:
 *
 * ------------------------------------------------------------------------------------------
 */
static int HandleMultipleValue(char *value, int(*cback)(const char *v, const char *p, lsParmguardCtxt *ctxt), 
					void *op1, void *op2) 
{
	int ret;

	/* scan the 'value' and call the callback for each word. */
	char *v = value;
	while (v && *v) {
		char *e = strchr(v,',');
		if (e) *e = '\0';

		if (ret = (*cback)(v,op1,op2))
			return ret;

		if (e) { 
			*e = '=';
			v = e+1;
		}
		else v = NULL;
	}

	return 0;	/* OK */
}

static int HandleBool(const char *value, int *bval)
{
	int	ival = atoi(value);

	if (ival == 0 || ival == 1) { *bval = ival; return 0; }

	if (!strcmp(value,"on") || 
	    !strcmp(value,"true")) { *bval = 1; return 0; }

	if (!strcmp(value,"off") || 
	    !strcmp(value,"false")) { *bval = 0; return 0; }

	return 1;	/* wrong value */
}

static int HandleMValueCBack(const char *value, const char *parm, lsParmguardCtxt *ctxt)
{
	int error = 0;

	if (!strcmp(parm,"undefined_url_action")) {
		if (!strcmp(value,"accept")) 		{ ctxt->undef_url_action_reject = 0; }
		else if (!strcmp(value,"reject")) 	{ ctxt->undef_url_action_reject = 1; }
		else if (!strcmp(value,"log")) 		{ ctxt->undef_url_action_log    = 1; }
		else if (!strcmp(value,"setenv")) 	{ ctxt->undef_url_action_setenv = 1; }
		else error = 1;
	}
	else if (!strcmp(parm,"undefined_parm_action")) {
		if (!strcmp(value,"accept")) 		{ ctxt->undef_parm_action_reject = 0; }
		else if (!strcmp(value,"reject")) 	{ ctxt->undef_parm_action_reject = 1; }
		else if (!strcmp(value,"clear"))	{ ctxt->undef_parm_action_clear  = 1; }
		else if (!strcmp(value,"log")) 		{ ctxt->undef_parm_action_log    = 1; }
		else if (!strcmp(value,"setenv")) 	{ ctxt->undef_parm_action_setenv = 1; }
		else error = 1;
	}
	else if (!strcmp(parm,"illegal_parm_action")) {
		if (!strcmp(value,"accept")) 		{ ctxt->illegal_parm_action_reject = 0; }
		else if (!strcmp(value,"reject")) 	{ ctxt->illegal_parm_action_reject = 1; }
		else if (!strcmp(value,"clear"))	{ ctxt->illegal_parm_action_clear  = 1; }
		else if (!strcmp(value,"log")) 		{ ctxt->illegal_parm_action_log    = 1; }
		else if (!strcmp(value,"setenv")) 	{ ctxt->illegal_parm_action_setenv = 1; }
		else error = 1;
	}
	else {
		(*perr)("Internal error: '%s' global parameter not handled",parm);
	}

	if (error) {
		(*perr)("Incorrect value: '%s' for Global Parameter '%s'",value,parm);
		return 1;
	}

	return 0;	/* OK */
}

static int HandleGlobal(guard_ctxt_t *ctxt, xmlDocPtr pDoc, xmlNodePtr gnode)
{
	xmlChar *name  = xmlGetProp(gnode,"name");
	xmlChar *value = xmlGetProp(gnode,"value");
	assert(name && value);	/* DTD assumed */

	if (!xmlStrcmp(name,(const xmlChar *)"http_error_code")) {
		ctxt->global_ctxt.http_error_code = atoi((const char *)value);
		if (ctxt->global_ctxt.http_error_code < 0) 
			goto badparm;
	}
	else if (!xmlStrcmp(name,(const xmlChar *)"catch_all_pass")) {
		(*perr)("'catch_all_pass' is obsolete and should be replaced by 'undefined_url_action'");
		if (HandleBool((const char *)value,&ctxt->global_ctxt.undef_url_action_reject)) 
			goto badparm;
	}
	else if (!xmlStrcmp(name,(const xmlChar *)"catch_all_log")) {
		(*perr)("'catch_all_log' is obsolete and should be replaced by 'undefined_url_action'");
		if (HandleBool((const char *)value,&ctxt->global_ctxt.undef_url_action_log)) 
			goto badparm;
	}
	else if (!xmlStrcmp(name,(const xmlChar *)"undefined_url_action")) {
		ctxt->global_ctxt.undef_url_action_log = ctxt->global_ctxt.undef_url_action_setenv = 0;
		if (HandleMultipleValue(value,HandleMValueCBack,"undefined_url_action",&ctxt->global_ctxt))
			goto badparm;
	}
	else if (!xmlStrcmp(name,(const xmlChar *)"illegal_parm_action")) {
		ctxt->global_ctxt.illegal_parm_action_log = ctxt->global_ctxt.illegal_parm_action_setenv = 0;
		if (HandleMultipleValue(value,HandleMValueCBack,"illegal_parm_action",&ctxt->global_ctxt))
			goto badparm;
	}
	else if (!xmlStrcmp(name,(const xmlChar *)"undefined_parm_action")) {
		ctxt->global_ctxt.undef_parm_action_log = ctxt->global_ctxt.undef_parm_action_setenv = 0;
		if (HandleMultipleValue(value,HandleMValueCBack,"undefined_parm_action",&ctxt->global_ctxt))
			goto badparm;
	}
	else if (!xmlStrcmp(name,(const xmlChar *)"scan_all_parm")) {
		if (HandleBool((const char *)value,&ctxt->global_ctxt.scan_all_parm)) 
			goto badparm;
	}
	else {
		(*perr)("Unsupported paramater: '%s'",name);
		assert(0);	/* DTD assumed */
	}

	return 0;

badparm:
	(*perr)("Incorrect value: '%s' for Global Parameter '%s'",value,name);
	return 1;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Main function. Parses the in-memory XML conf file and builds an internal representation.
 *
 * ------------------------------------------------------------------------------------------
 */
int BuildRules(guard_ctxt_t *ctxt, xmlDocPtr pDoc)
{
	/* Run through the Document and build the internal rules: */
	xmlNodePtr node = xmlDocGetRootElement(pDoc);
	xmlNodePtr cur;

	if (!node) {
		(*perr)("Empty Rule File ?");
		return 0;
	}

	/* Loop on the children: */
	cur = node->xmlChildrenNode;
	while (cur) {
		if (cur->type != XML_ELEMENT_NODE) {
			cur = cur->next;
			continue;
		}

		if (!xmlStrcmp(cur->name,(const xmlChar *)"usertype")) {
			int status = BuildUserType(ctxt,pDoc,cur);
			if (status) return status;
		}
		else if (!xmlStrcmp(cur->name,(const xmlChar *)"url")) {
			int status = BuildURLRule(ctxt,pDoc,cur);
			if (status) return status;
		}
		else if (!xmlStrcmp(cur->name,(const xmlChar *)"global")) {
			int status = HandleGlobal(ctxt,pDoc,cur);
			if (status) return status;
		}
		else assert(0);	/* DTD assumed */

		cur = cur->next;
	}

	return 0;
}

/*
 * ==========================================================================================
 *
 * Structural functions that allocate and manage with the internal representation of rules.
 *
 * ==========================================================================================
 */
/*
 * ------------------------------------------------------------------------------------------
 *
 * Allocate a new rule definition structure:
 *
 * ------------------------------------------------------------------------------------------
 */
guard_rule_t *AddRule(guard_ctxt_t *ctxt, const char *server, int port, const char *path)
{
	guard_rule_t *p = APR_PALLOC(ctxt->pool,sizeof(guard_rule_t));
	ap_regex_t *reg = NULL;
	assert(p);
	memset(p,0,sizeof(guard_rule_t));
	if (server) p->m_server = (const char *)APR_PSTRDUP(ctxt->pool,server);
	p->m_port = port;
	if (path) p->m_path = (const char *)APR_PSTRDUP(ctxt->pool,path);
	p->m_parm = NULL;
	NEXT(p) = ctxt->p_guard_rule;

	/* compile the regexp: */
	if (!(reg = AP_PREGCOMP(ctxt->pool,path,AP_REG_EXTENDED|AP_REG_NOSUB))) {
		(*perr)("Bad RE for <match> tag: %s",path);
		return NULL;
	}
	p->m_regexp = reg;

	ctxt->p_guard_rule = p;

	return ctxt->p_guard_rule;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Allocate a new parm definition structure:
 *
 * ------------------------------------------------------------------------------------------
 */
guard_parm_t *AddParm(apr_pool_t *pool, guard_rule_t *rule, const char *parm)
{
	/* Create a new structure and insert it at the head of the list: */
	guard_parm_t	*p = APR_PALLOC(pool,sizeof(guard_parm_t));
	assert(p);
	memset(p,0,sizeof(guard_parm_t));
	NAME(p) = (const char *)APR_PSTRDUP(pool,parm);
	assert(NAME(p));
	assert(rule);
	NEXT(p) = rule->m_parm;

	rule->m_parm = p;
	
	return p;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Allocate a new type definition structure:
 *
 * ------------------------------------------------------------------------------------------
 */
guard_type_t *AddType(guard_ctxt_t *ctxt, const char *name)
{
	guard_type_t *p = APR_PALLOC(ctxt->pool,sizeof(guard_type_t));

	/* Create a new structure and insert it at the head of the list: */
	if (g_debuglevel)
		(*perr)("AddType: ctxt=0x%x, name=%s",ctxt,name);

	assert(p);
	memset(p,0,sizeof(guard_type_t));
	NAME(p) = (const char *)APR_PSTRDUP(ctxt->pool,name);
	assert(NAME(p));
	p->m_pattr = NULL;
	NEXT(p) = ctxt->p_guard_type;

	ctxt->p_guard_type = p;

	return ctxt->p_guard_type;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Allocate a new base type definition structure:
 *
 * ------------------------------------------------------------------------------------------
 */
guard_type_t *AddBaseType(guard_ctxt_t *ctxt, const char *name, CONSTRAINT_BASE_TYPE bt)
{
	/* Create a new structure and insert it at the head of the list: */
	guard_type_t *t = AddType(ctxt,name);
	t->m_bt = bt;

	return t;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Allocate a new user type definition structure:
 *
 * ------------------------------------------------------------------------------------------
 */
guard_type_t *AddUserType(guard_ctxt_t *ctxt, const char *name)
{
	/* Create a new structure and insert it at the head of the list: */
	guard_type_t *t = AddType(ctxt,name);
	t->m_userdefined = 1;

	return t;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Allocate a structure to add a new attribute to a type:
 *
 * ------------------------------------------------------------------------------------------
 */
void AddTypeAttr(apr_pool_t *pool, guard_type_t *t, const char *name, CONSTRAINT_TYPE_ATTR ct)
{
	guard_type_attr_t *p = APR_PALLOC(pool,sizeof(guard_type_attr_t));
	assert(p);
	memset(p,0,sizeof(guard_type_attr_t));
	NAME(p) = (const char *)APR_PSTRDUP(pool,name);
	NEXT(p) = t->m_pattr;
	p->m_ct = ct;
	t->m_pattr = p;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Allocate a structure to add a new value to a rule parm:
 *
 * ------------------------------------------------------------------------------------------
 */
guard_aval_t *AddValue(apr_pool_t *pool, guard_aval_t **m_value, guard_type_attr_t *attr, const char *value)
{
	guard_aval_t *v = APR_PALLOC(pool,sizeof(guard_aval_t));
	char *endptr;
	assert(v);
	memset(v,0,sizeof(guard_aval_t));
	v->m_svalue = (const char *)APR_PSTRDUP(pool,value);
	assert(v->m_svalue);
	v->m_lvalue = atol(value);
	v->m_fvalue = strtod(value,&endptr);
	if (g_debuglevel)
		(*perr)("AddValue: value=%s, m_fvalue=%f",value,v->m_fvalue);
	v->m_attr = attr;
	NEXT(v) = *m_value;
	*m_value = v;

	return v;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Search functions:
 *
 * ------------------------------------------------------------------------------------------
 */
guard_type_t *GetTypeList(guard_ctxt_t *ctxt) 
{
	return ctxt->p_guard_type;
}

guard_rule_t *GetRuleList(guard_ctxt_t *ctxt)
{
	return ctxt->p_guard_rule;
}

guard_type_t *SearchType(guard_type_t *tlist, const char *name)
{
	while (tlist) {
		if (!strcmp(NAME(tlist),name)) return tlist;
		tlist = NEXT(tlist);
	}

	return NULL;
}

guard_type_attr_t *SearchAttr(guard_type_t *t, const char *name)
{
	guard_type_attr_t *a;

	/* For user defined type, the base type is taken in account: */
	if (t->m_userdefined)
		t = t->m_basetype;

	a = t->m_pattr;
	while (a) {
		if (!strcmp(NAME(a),name)) return a;
		a = NEXT(a);
	}

	return NULL;
}

guard_parm_t *SearchParm(guard_rule_t *rule, const char *name)
{
	guard_parm_t *p = rule->m_parm;

	while (p) {
		if (!strcmp(NAME(p),name)) return p;
		p = NEXT(p);
	}

	return NULL;
}

/*
 * ------------------------------------------------------------------------------------------
 *
 * Dump the rules in memory (for debugging purpose)
 *
 * ------------------------------------------------------------------------------------------
 */
void DumpRules(guard_ctxt_t *ctxt)
{
	guard_rule_t *r; 
	guard_type_t *t;

	if (!ctxt) return;

	r = ctxt->p_guard_rule;
	t = ctxt->p_guard_type;

	if (g_debuglevel)
		(*perr)("DumpRules: ctxt=0x%x",ctxt);

	/* Dump the types: */
	while (t) {
		(*perr)("Dump of type '%s':\n",NAME(t));
		if (t->m_userdefined) {
			guard_aval_t	*aval = t->m_value;

			while (aval) {
				(*perr)("\tattribute: %s, value=%s\n",NAME(aval->m_attr),aval->m_svalue);
				aval = NEXT(aval);
			}
		}
		else {
			guard_type_attr_t *a = t->m_pattr;
			while (a) {
				(*perr)("\towns attribute: '%s'\n",NAME(a));
				a = NEXT(a);
			}
		}

		t = NEXT(t);
	}

	/* Dump the rules: */
	while (r) {
		guard_parm_t *p = r->m_parm;
		(*perr)("Dump of rule for server %s:%d, URL:%s\n",r->m_server,r->m_port,r->m_path);
		while (p) {
			guard_aval_t	*aval = p->m_value;
			(*perr)("\tparm %s\n",NAME(p));
			(*perr)("\t\ttype: %s\n",NAME(p->m_type));

			while (aval) {
				(*perr)("\t\tattribute: %s, value=%s\n",NAME(aval->m_attr),aval->m_svalue);
				aval = NEXT(aval);
			}

			p = NEXT(p);
		}
		r = NEXT(r);
	}
}

/*
 * $Log:$
 */

/* EOF */
