/*
 * ==========================================================================================
 * rules.h
 *
 * Date: 10 May 2004 
 * Author: Jerome Delamarche
 * Version: 1.3 
 *
 * ------------------------------------------------------------------------------------------
 * CHANGES:
 *
 * ------------------------------------------------------------------------------------------
 *
 * TODO:
 *
 * ------------------------------------------------------------------------------------------
 *
 * Rules internal structure:
 *
 * ==========================================================================================
 */
#ifndef _mod_parmguard_rules_h_
#define _mod_parmguard_rules_h_

#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "http_core.h"
#include <libxml/SAX.h>

#define NAME(x)	(x)->m_name
#define NEXT(x) (x)->m_next

/*
 * Structure that describes the engine parameters:
 * There is a "global" structure that matches the global parameter values.
 * Other structures can exist and attached to the URIs. They overwrite the global structure.
 */
typedef struct {
	int http_error_code;			/* value to be returned when request is denied by mod_parmguard */
	int undef_url_action_reject;		/* if true, rejects the requests not 'parmguarded' which have args */
	int undef_url_action_log;		/* if true, log all the requests not 'parmguarded' which have args */
	int undef_url_action_setenv;		/* if true, set an environment variable for all the requests not 'parmguarded' which have args */
	int illegal_parm_action_reject;		/* if true, invalid parm rejects the request */
	int illegal_parm_action_clear;		/* if true, invalid parm are set to "" - request not rejected */
	int illegal_parm_action_filter;		/* if true, invalid chars in parms are removed - request not rejected */
	int illegal_parm_action_log;		/* if true, log an error for each invalid parm */
	int illegal_parm_action_setenv;		/* if true, set an environment variable for each invalid parm */
	int undef_parm_action_reject;		/* if true, rejects request with undefined parm constraint */
	int undef_parm_action_clear;		/* if true, undef parm are set to "" - request not rejected */
	int undef_parm_action_log;		/* if true, all undefined parm are logged */
	int undef_parm_action_setenv;		/* if true, an environment variable is set for each undefined parm */
	int scan_all_parm; 			/* if true, do not stop after the 1st invalid parm */
} lsParmguardCtxt;

/*
 * In order to enhance the parameters checkings, the constraints name are
 * mapped to an integer value:
 */
typedef enum _constraint_type_attr_t {
	/* string type */
	/* note: sometimes maxlen and minlen may be susbstituted by a RE using charclass */
	CONSTRAINT_MAXLEN = 1,
	CONSTRAINT_MINLEN,
	CONSTRAINT_CHARCLASS,
	/* integer and decimal types: */
	CONSTRAINT_MAXVAL,
	CONSTRAINT_MINVAL,
	/* enum type: */
	CONSTRAINT_MULTIPLE,
	CONSTRAINT_OPTION,
} CONSTRAINT_TYPE_ATTR;

/*
 * Base types are also mapped into an enum value:
 */
typedef enum _base_type_t {
	CONSTRAINT_STRING = 1,
	CONSTRAINT_INTEGER,
	CONSTRAINT_DECIMAL,
	CONSTRAINT_ENUM,
} CONSTRAINT_BASE_TYPE;

/*
 * Definition of formal attributes:
 */
typedef struct _guard_type_attr_t {
	const char			*m_name;	/* attribute name */
	CONSTRAINT_TYPE_ATTR		m_ct;		/* equivalent value */
	struct _guard_type_attr_t	*m_next;	/* next attribute */
} guard_type_attr_t;

/*
 * Definition of attributes values:
 */
typedef struct _guard_aval_t {
	guard_type_attr_t		*m_attr;
	const char			*m_svalue;	/* string representation of the value */
	long				m_lvalue;	/* long value */
	double				m_fvalue;	/* float value */
	ap_regex_t			*m_regexp;	/* compiled regexp for 'charclass' constraint */
	struct _guard_aval_t		*m_next;
} guard_aval_t;

/*
 * Definition of a parameter type:
 */
typedef struct _guard_type_t {
	const char			*m_name;	/* type name */
	guard_type_attr_t		*m_pattr;	/* ptr to the first type attribute (used for conf file validation) */
	int				m_userdefined;	/* 1 if user defined ! */
	CONSTRAINT_BASE_TYPE		m_bt;		/* base type value */
	struct _guard_type_t		*m_basetype;	/* base type for user defined type */
	guard_aval_t			*m_value;	/* list of attribute value for user defined type */
	struct _guard_type_t		*m_next;	/* next type */
} guard_type_t;

/*
 * Definition of a parameter constraint:
 */
typedef struct _guard_parm_t {
	const char			*m_name;
	guard_type_t			*m_type;
	guard_aval_t			*m_value;	/* list of attribute value */
	struct _guard_parm_t		*m_next;
} guard_parm_t;

/*
 * Structure of a rule:
 */
typedef struct _guard_rule_t {
	const char 			*m_server;	/* server name: www.server.com , if NULL, means *ANY* server */
	int				m_port;		/* HTTP Port, if NULL, means *ANY* port  */
	const char			*m_path;	/* everything before the parms, accepts metachars, NULL means *ANY* URL */
	ap_regex_t			*m_regexp;	/* compiled path */
	guard_parm_t			*m_parm;	/* first constraint on parms */
	struct _guard_rule_t		*m_next;	/* ptr to next rule */
	lsParmguardCtxt			*m_ctxt;	/* NULL except if overwrite global parameters (unused) */
} guard_rule_t;

/* 
 * private configuration: 
 */
typedef struct {
	const char			*conffile;	/* XML configuration file name */
	guard_rule_t			*p_guard_rule;
	guard_type_t			*p_guard_type;
	lsParmguardCtxt			global_ctxt;
	apr_pool_t			*pool;		/* pool where the structure is allocated from */
} guard_ctxt_t;

/*
 * Exported global variables:
 */
extern int g_debuglevel;		
extern lsParmguardCtxt g_global_ctxt;

/*
 * Exported functions:
 */
extern guard_ctxt_t *InitRules(int debuglevel, const char *f, void (*errfunc)(const char *format, ...));
extern int BuildRules(guard_ctxt_t *cfg, xmlDocPtr pDoc);
extern void EndRules(guard_ctxt_t *ctxt);
extern void DumpRules(guard_ctxt_t *cfg);
extern int ApplyRules(void (*errfunc)(const char *format, ...), guard_ctxt_t *cfg, request_rec *r);
extern guard_rule_t *AddRule(guard_ctxt_t *cfg, const char *server, int port, const char *path);
extern guard_parm_t *AddParm(apr_pool_t *pool, guard_rule_t *rule, const char *parm);
extern guard_type_t *AddType(guard_ctxt_t *cfg, const char *name);
extern guard_type_t *AddBaseType(guard_ctxt_t *cfg, const char *name, CONSTRAINT_BASE_TYPE bt);
extern guard_type_t *AddUserType(guard_ctxt_t *cfg, const char *name);
extern guard_aval_t *AddValue(apr_pool_t *pool, guard_aval_t **m_value, guard_type_attr_t *attr, const char *value);
extern void AddTypeAttr(apr_pool_t *pool, guard_type_t *t, const char *name, CONSTRAINT_TYPE_ATTR ct);
extern guard_type_t *SearchType(guard_type_t *tlist, const char *name);
extern guard_type_attr_t *SearchAttr(guard_type_t *t, const char *name);
extern guard_parm_t *SearchParm(guard_rule_t *rule, const char *name);
extern guard_type_t *GetTypeList(guard_ctxt_t *cfg);
extern guard_rule_t *GetRuleList(guard_ctxt_t *cfg);

#endif /* _mod_parmguard_rules_h_ */

/* EOF */
